@@ -1,12 +1,12 @@
-import cv2
-import json
-import numpy as np
 import logging
 from typing import Callable
-from yupi.tracking.algorithms import Algorithm, resize_frame
+from pathlib import Path
+import cv2
+import numpy as np
+from yupi.tracking.algorithms import TrackingAlgorithm, resize_frame
 from yupi.affine_estimator import get_affine
 from yupi.trajectory import Trajectory
-from yupi.analyzing.reference import add_dynamic_reference
+from yupi.analyzing.transformations import add_dynamic_reference
 from yupi.tracking.undistorters import Undistorter
 
 
@@ -14,40 +14,43 @@ class ROI():
     """
     Region of interest.
 
-    Region that can be tracked by the algorithms throughout the sequence of
-    image frames.
+    Region that can be tracked by the algorithms throughout the sequence
+    of image frames.
 
     Parameters
     ----------
     size : tuple of float
         Size of the region of interest.
 
-        If both tuple's values are grater than 1 then they are rounded and
-        taken as pixels. Otherwise, if both values are less than 1, the
-        size is taken relative to the video frame size.
+        If both tuple's values are grater than 1 then they are rounded
+        and taken as pixels. Otherwise, if both values are less than 1,
+        the size is taken relative to the video frame size.
     init_mode : str, optional
-        ROI's initialization mode. (Default is 'manual').
+        ROI's initialization mode, by default 'manual'.
 
         Defines the way ROI initial position is setted.
 
-        The ``init_mode`` parameter can be manual or center. These modes are
-        stored in ``ROI.MANUAL_INIT_MODE`` and ``ROI.CENTER_INIT_MODE``.
+        The ``init_mode`` parameter can be manual or center. These
+        modes are stored in ``ROI.MANUAL_INIT_MODE`` and
+        ``ROI.CENTER_INIT_MODE``.
     scale : float, optional
         Scale of the sample frame to set ROI initial position if
-        ``init_method`` is set to ``'manual'``. (Default is 1).
+        ``init_method`` is set to ``'manual'``, by default 1.
 
     Attributes
     ----------
     width : float
         Width of the ROI.
 
-        If the width value is between 0 and 1 then this is taken relative
-        to the frames. Otherwise it is a rounded value and taken as pixels.
+        If the width value is between 0 and 1 then this is taken
+        relative to the frames. Otherwise it is a rounded value and
+        taken as pixels.
     height : float
         Height of the ROI.
 
-        If the height value is between 0 and 1 then this is taken relative
-        to the frames. Otherwise it is a rounded value and taken as pixels.
+        If the height value is between 0 and 1 then this is taken
+        relative to the frames. Otherwise it is a rounded value and
+        taken as pixels.
     init_mode : str
         ROI's initialization mode.
     scale : float
@@ -64,9 +67,11 @@ class ROI():
     ValueError
         If any size value is negative.
     ValueError
-        If one of the size value is grater than 1 and the other is less than 1.
+        If one of the size value is grater than 1 and the other is less
+        than 1.
     ValueError
-        If ROI initialization mode is neither ``'manual'`` or ``'center'``.
+        If ROI initialization mode is neither ``'manual'`` or
+        ``'center'``.
     """
 
     MANUAL_INIT_MODE = 'manual'
@@ -83,12 +88,15 @@ class ROI():
         # TODO: check for a more pythonic way to do this comprobation
         if (size[0] < 1 and size[1] > 1) or \
            (size[0] > 1 and size[1] < 1):
-            raise ValueError(size)  # TODO: write error message here
+            raise ValueError('Size values must be between 0 and 1 both or '
+                             'integers greater than 0 both')
 
-        if init_mode != ROI.CENTER_INIT_MODE and \
-           init_mode != ROI.MANUAL_INIT_MODE:
+        if init_mode not in (ROI.CENTER_INIT_MODE, ROI.MANUAL_INIT_MODE):
             raise ValueError(f"ROI '{init_mode}' initialization mode unknown")
-        
+
+        if scale < 0:
+            raise ValueError('ROI scale must be non negative')
+
         self.width, self.height = size
         self.init_mode = init_mode
         self.scale = scale
@@ -96,7 +104,7 @@ class ROI():
         self.__cXY = None, None
         self.__global_height, self.__global_width = None, None
 
-    # this repr could change
+    # This repr could change
     def __repr__(self):
         return 'ROI: size=({}, {}) init_mode={} scale={}' \
             .format(self.width, self.height, self.init_mode, self.scale)
@@ -116,29 +124,30 @@ class ROI():
             Center of the ROI.
         """
 
-        # get the centroid refered to the roi
-        cX_roi, cY_roi = centroid
+        # Get the centroid refered to the roi
+        if centroid is not None:
+            cX_roi, cY_roi = centroid
 
-        # get the centroid refered to the full image
-        cX = self.__prev_cXY[0] - int(self.width/2) + cX_roi
-        cY = self.__prev_cXY[1] - int(self.height/2) + cY_roi
-
-        cX = min(cX, self.__global_width)
-        cX = max(cX, 0)
-        cY = min(cY, self.__global_height)
-        cY = max(cY, 0)
-       
-        self.__cXY = cX, cY
+            # Get the centroid refered to the full image
+            cX = self.__prev_cXY[0] - int(self.width/2) + cX_roi
+            cY = self.__prev_cXY[1] - int(self.height/2) + cY_roi
 
+            cX = min(cX, self.__global_width)
+            cX = max(cX, 0)
+            cY = min(cY, self.__global_height)
+            cY = max(cY, 0)
 
+            self.__cXY = cX, cY
 
+        else:
+            self.__cXY = self.__prev_cXY
 
     def _get_bounds(self, prev: bool = False) -> tuple:
         """
         ROI's bounds.
 
-        Calculates the ROI's bounds according to its center, width, height and
-        the global bounds.
+        Calculates the ROI's bounds according to its center, width,
+        height and the global bounds.
 
         Parameters
         ----------
@@ -175,7 +184,8 @@ class ROI():
 
         Parameters:
         frame : np.ndarray
-            Frame used as reference to initialize ROI position at its center.
+            Frame used as reference to initialize ROI position at its
+            center.
 
         Returns
         -------
@@ -184,13 +194,12 @@ class ROI():
         """
 
         self.__global_height, self.__global_width = frame.shape[:2]
-        self.__cXY = int(self.__global_width/2), int(self.__global_height/2)
+        self.__cXY = self.__global_width // 2, self.__global_height // 2
         return self.__cXY
 
-    # TODO: check for 'win2_name' utility. Maybe it it should be 'ROI' as
-    # default so there is no need to pass it as a parameter
-    def _manual_init(self, frame: np.ndarray, name: str,
-                      win2_name: str = 'ROI') -> tuple:
+    # TODO: check for 'win2_name' utility. Maybe it should be 'ROI' as
+    # Default so there is no need to pass it as a parameter
+    def _manual_init(self, frame: np.ndarray, name: str) -> tuple:
         """
         Initialize ROI using manual initialization mode.
 
@@ -214,18 +223,17 @@ class ROI():
         frame_ = resize_frame(frame, scale=self.scale)
         cv2.imshow(win1_name, frame_)
 
-        # callback handler to manually set the roi
+        # Callback handler to manually set the roi
         def on_click(event, x, y, flags, param):
 
             if event == cv2.EVENT_LBUTTONDOWN:
-                # global roi center coordinates
+                # Global roi center coordinates
                 self.__cXY = int(x / self.scale), int(y / self.scale)
 
-                # copy of true frame and its resized version
-                img = frame.copy()
+                # Copy of true frame and its resized version
                 img_ = frame_.copy()
 
-                # draw a circle in the selected pixel
+                # Draw a circle in the selected pixel
                 cv2.circle(img_, (x, y), 3, (0, 255, 255), 1)
                 xmin, xmax, ymin, ymax = self._get_bounds()
                 pt1 = (int(xmin * self.scale), int(ymin * self.scale))
@@ -233,25 +241,12 @@ class ROI():
                 cv2.rectangle(img_, pt1, pt2, (0, 255, 255), 1)
                 cv2.imshow(win1_name, img_)
 
-                # get roi in the full size frame
-
-                # cv2.circle(img_, (x, y), 3, (0, 255, 255), 1)
-                # roi = self._crop(img_)
-
-                # roi padding just to display the new window
-                # padL, padR = np.hsplit(np.zeros_like(roi), 2)
-                # roi_ = np.hstack([padL, roi, padR])
-                # cv2.imshow(win2_name, roi_)
-
                 logging.info('ROI initialized, now press any key to continue')
 
         cv2.setMouseCallback(win1_name, on_click)
         cv2.waitKey(0)
         return self.__cXY
 
-    # TODO: check for 'name' utility. It is only use for the return message
-    # I think this method should only return True/False and then handle the
-    # error in the tracking scenario
     def _check_roi_init(self, name: str) -> bool:
         """
         Checks for ROI initialization.
@@ -286,8 +281,8 @@ class ROI():
         first_frame : np.ndarray
             First frame of the video.
 
-            If ROI's initialization mode is set to ``'manual'`` this frame
-            will be shown to select the tracking object center.
+            If ROI's initialization mode is set to ``'manual'`` this
+            frame will be shown to select the tracking object center.
 
         Returns
         -------
@@ -329,44 +324,44 @@ class ROI():
         """
 
         self.__global_height, self.__global_width = frame.shape[:2]
-        # bounds of the roi
+        # Bounds of the roi
         xmin, xmax, ymin, ymax = self._get_bounds(prev)
         window = frame[ymin:ymax, xmin:xmax, :]
-        return window
+        return window.copy()
 
 
 class ObjectTracker():
     """
-    Tracks an object inside a ROI according to an algorithm.
+    Tracks an object inside a ROI according to a tracking algorithm.
 
     Parameters
     ----------
     name : str
         Name of the tracked object.
-    algorithm : Algorithm
+    algorithm : TrackingAlgorithm
         Algorithm used to track the object.
     roi : ROI
         Region of interest where the object will be tracked.
     preprocessing : Callable[[np.ndarray], np.ndarray], optional
-        Preprocessing function aplied to the frame before being used by the
-        algorithm 
+        Preprocessing function aplied to the frame before being used by
+        the algorithm.
 
     Attributes
     ----------
     name : str
         Name of the tracked object.
-    algorithm : Algorithm
+    algorithm : TrackingAlgorithm
         Algorithm used to track the object.
     roi : ROI
         Region of interest where the object will be tracked.
     history : list of tuple
         ROI's position in every frame of the video.
     preprocessing : Callable[[np.ndarray], np.ndarray]
-        Preprocessing function aplied to the frame before being used by the
-        algorithm
+        Preprocessing function aplied to the frame before being used by
+        the algorithm.
     """
 
-    def __init__(self, name: str, algorithm: Algorithm, roi: ROI,
+    def __init__(self, name: str, algorithm: TrackingAlgorithm, roi: ROI,
                  preprocessing: Callable[[np.ndarray], np.ndarray] = None):
         self.name = name
         self.roi = roi
@@ -387,23 +382,22 @@ class ObjectTracker():
         Parameters
         ----------
         frame : np.ndarray
-            Frame used by the algorithm to detect the tracked object's new
-            center.
+            Frame used by the algorithm to detect the tracked object's
+            new center.
         """
-        # get only the ROI from the current frame
-        window = self.roi._crop(frame)
 
-        # Preprocess the image
-        if self.preprocessing is not None:
-            window = self.preprocessing(window)
+        # Get only the ROI from the current frame
+        roi_bound = self.roi._get_bounds()
 
-        # detect the object using the tracking algorithm
-        self.mask, centroid = self.algorithm.detect(window)
+        # Detect the object using the tracking algorithm
+        self.mask, centroid = self.algorithm.detect(frame,
+                                                    roi_bound,
+                                                    self.preprocessing)
 
-        # update the roi center using current ant coordinates
+        # Update the roi center using current ant coordinates
         self.roi._recenter(centroid)
 
-        # update data
+        # Update data
         self.history.append(self.roi._ROI__cXY)
 
 
@@ -414,12 +408,14 @@ class CameraTracker():
     Parameters
     ----------
     roi : ROI
-        Region of interest where the background changes will be detected.
+        Region of interest where the background changes will be
+        detected.
 
     Attributes
     ----------
     roi : ROI
-        Region of interest where the background changes will be detected.
+        Region of interest where the background changes will be
+        detected.
     history : list of tuple
         ROI's position in every frame of the video.
     """
@@ -428,13 +424,14 @@ class CameraTracker():
         self.history = []
         self.mse = []
         self.roi = roi
+        self.features = None
 
     def _init_roi(self, prev_frame: np.ndarray) -> bool:
         return self.roi._initialize('Camera', prev_frame)
 
-    # track the floor
+    # Track the floor
     def _track(self, prev_frame: np.ndarray, frame: np.ndarray,
-              ignored_regions: list) -> bool:
+               ignored_regions: list) -> bool:
         """
         Tracks the camera movements according to the changing background
         inside the ROI.
@@ -446,25 +443,30 @@ class CameraTracker():
         igonerd_regions : list of tuple
             Tracked object's boundaries.
 
-            Tracked object's does not form part of the background so they
-            should be ignored.
+            Tracked object's does not form part of the background so
+            they should be ignored.
 
         Returns
         -------
         bool
-            Whether or not good points were found or sucessfully tracked.
+            Whether or not good points were found or sucessfully
+            tracked.
         """
+
         # Initialize a mask of what to track
         h, w = frame.shape[:2]
         mask = 255 * np.ones((h, w), dtype=np.uint8)
 
-        # mask pixeles inside every ROIs
+        # Mask pixeles inside every ROIs
         for x0, xf, y0, yf in ignored_regions:
             mask[y0:yf, x0:xf] = 0
 
-        p_good, aff_params, err = get_affine(prev_frame, frame,
-                                             self.roi._get_bounds(),
-                                             mask)
+        p_good, aff_params, err = get_affine(
+            img1=prev_frame,
+            img2=frame,
+            region=self.roi._get_bounds(),
+            mask=mask
+        )
         self.features = p_good[1:]
 
         if err is None:
@@ -485,20 +487,20 @@ class TrackingScenario():
     object_trackers : list
         Trackers of all the objects.
     camera_tracker : CameraTracker
-        Tracker used to detect camera movements. (Default is None).
+        Tracker used to detect camera movements, by default None.
     undistorter : Undistorter
-        Undistorted used to correct each video frame. (Default is None).
+        Undistorted used to correct each video frame, by default None.
     preview_scale : float
-        Scale of the video preview. (Default is 1.0).
+        Scale of the video preview, by default 1.0.
     auto_mode : bool
-        If True the video is processed auomtically otherwise it's processed
-        manually. (Default is True).
+        If True the video is processed auomtically otherwise it's
+        processed manually, by default True.
 
         If the video is processed manually, pressing ``ENTER`` key is
         necessary in every frame to continue.
 
-        This mode can be changed in the middle of the processing by pressing
-        ``M`` key.
+        This mode can be changed in the middle of the processing by
+        pressing ``M`` key.
 
     Attributes
     ----------
@@ -511,14 +513,14 @@ class TrackingScenario():
     preview_scale : float
         Scale of the video preview.
     auto_mode : bool
-        If True the video is processed auomtically otherwise it's processed
-        manually. (Default is True).
+        If True the video is processed auomtically otherwise it's
+        processed manually, by default True.
 
         If the video is processed manually, pressing ``Enter`` key is
         necessary in every frame to continue.
 
-        This mode can be changed in the middle of the processing by pressing
-        ``M`` key.    
+        This mode can be changed in the middle of the processing by
+        pressing ``M`` key.
     """
 
     def __init__(self, object_trackers: list,
@@ -530,27 +532,39 @@ class TrackingScenario():
         self.camera_tracker = camera_tracker
         self.undistorter = undistorter
         self.preview_scale = preview_scale
-        self.auto_mode = True
+        self.auto_mode = auto_mode
         self._enabled = True
         self._iteration_counter = 0
 
+        self.video_path = None
+        self.cap = None
+        self.frame_count = None
+        self.fps = None
+        self.w = None
+        self.h = None
+        self.dim = None
+        self.first_frame = None
+        self.last_frame = None
+        self.prev_frame = None
+
     def _digest_video_path(self, video_path):
-        # TODO: Validate the path
+        if not Path.exists(Path(video_path)):
+            raise ValueError(f"Path '{video_path}' does not exists")
         self.video_path = video_path
 
-        # create capture object
+        # Create capture object
         self.cap = cv2.VideoCapture(video_path)
 
-        # total number of frames in the video file
+        # Total number of frames in the video file
         self.frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
 
-        # frames per seconds
+        # Frames per seconds
         self.fps = self.cap.get(cv2.CAP_PROP_FPS)
 
-        # frame width
+        # Frame width
         self.w = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
 
-        # frame height
+        # Frame height
         self.h = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
 
         self.dim = (self.w, self.h)
@@ -563,10 +577,10 @@ class TrackingScenario():
         return frame
 
     def _show_frame(self, frame, show_frame_id=True):
-        # cXY, region, features, frame_numb, mask
+        # CXY, region, features, frame_numb, mask
         frame = frame.copy()
 
-        # draw region in which features are detected
+        # Draw region in which features are detected
         if self.camera_tracker:
             x0, xf, y0, yf = self.camera_tracker.roi._get_bounds()
 
@@ -576,7 +590,7 @@ class TrackingScenario():
 
             cv2.rectangle(frame, (x0, y0), (xf, yf), (0, 0, 255), 2)
             p2, p3 = self.camera_tracker.features
-            # draw detected and estimated features
+            # Draw detected and estimated features
             for p2_, p3_ in zip(p2, p3):
                 x2, y2 = np.rint(p2_).astype(np.int32)
                 x3, y3 = np.rint(p3_).astype(np.int32)
@@ -586,11 +600,12 @@ class TrackingScenario():
 
         for otrack in self.object_trackers:
             # TODO: Do this better:
-            # alter the blue channel in ant-related pixels
+            # Alter the blue channel in ant-related pixels
             window = otrack.roi._crop(frame, prev=True)
-            window[:, :, 0] = otrack.mask
+            if otrack.mask is not None:
+                window[:, :, 0] = otrack.mask
 
-            # draw a point over the roi center and draw bounds
+            # Draw a point over the roi center and draw bounds
             x1, x2, y1, y2 = otrack.roi._get_bounds()
             cv2.circle(frame, otrack.roi._ROI__cXY, 5, (255, 255, 255), -1)
             cv2.rectangle(frame, (x1, y1), (x2, y2), (0, 255, 255), 2)
@@ -609,18 +624,18 @@ class TrackingScenario():
 
         frame = resize_frame(frame, self.preview_scale)
         cv2.imshow('yupi processing window', frame)
-        # return frame
+        # Return frame
 
     def _first_iteration(self, start_in_frame):
         # Start processing frams at the given index
         if start_in_frame:
             self.first_frame = start_in_frame
-            self.cap.set(cv2.CAP_PROP_P_FRAMES, start_in_frame)
+            self.cap.set(cv2.CAP_PROP_POS_FRAMES, start_in_frame)
 
         # Capture the first frame to process
-        ret, prev_frame = self.cap.read()
+        _, prev_frame = self.cap.read()
 
-        # correct spherical distortion
+        # Correct spherical distortion
         self.prev_frame = self._undistort(prev_frame)
 
         # Initialize the roi of all the trackers
@@ -640,16 +655,13 @@ class TrackingScenario():
         return True
 
     def _keyboard_controller(self):
-        # keyboard events
+        # Keyboard events
         wait_key = 0 if not self.auto_mode else 10
 
         k = cv2.waitKey(wait_key) & 0xff
         if k == ord('m'):
             self.auto_mode = not self.auto_mode
 
-        if k == ord('s'):
-            self.__export_data__()
-
         elif k == ord('q'):
             self._enabled = False
 
@@ -657,14 +669,20 @@ class TrackingScenario():
             exit()
 
     def _regular_iteration(self):
-        # get current frame and ends the processing when no more frames are
+        # Get current frame and ends the processing when no more frames are
         # detected
+
+        frame_id = self._iteration_counter + self.first_frame
+        if self.last_frame is not None and frame_id >= self.last_frame:
+            return False, True
+
+
         ret, frame = self.cap.read()
         if not ret:
             logging.info('All frames were processed')
             return False, True
 
-        # correct spherical distortion
+        # Correct spherical distortion
         frame = self._undistort(frame)
 
         # ROI Arrays of tracking objects
@@ -678,22 +696,21 @@ class TrackingScenario():
 
         if self.camera_tracker:
             ret = self.camera_tracker._track(self.prev_frame, frame,
-                                                 roi_array)
-        frame_id = self._iteration_counter + self.first_frame
+                                             roi_array)
 
         if not ret:
             msg = f'CameraTracker - No matrix was estimated (Frame {frame_id})'
             logging.error(msg)
             return False, False
 
-        # display the full image with the ant in blue (TODO: Refactor this to
-        # make more general)
+        # Display the full image with the ant in blue (TODO: Refactor this to
+        # Make more general)
         self._show_frame(frame)
 
         for otrack in self.object_trackers:
             otrack.roi._ROI__prev_cXY = otrack.roi._ROI__cXY
 
-        # save current frame and ROI center as previous for next iteration
+        # Save current frame and ROI center as previous for next iteration
         self.prev_frame = frame.copy()
 
         # Call the keyboard controller to handle key interruptions
@@ -708,14 +725,13 @@ class TrackingScenario():
         self.cap.release()
         cv2.destroyAllWindows()
 
-
     def _tracker2trajectory(self, tracker, pix_per_m):
-        dt = self.fps
-        id = tracker.name
+        dt = 1/self.fps
+        traj_id = tracker.name
         x, y = map(list, zip(*tracker.history))
         x_arr = np.array(x) / pix_per_m
         y_arr = -1 * np.array(y) / pix_per_m
-        return Trajectory(x=x_arr, y=y_arr, dt=dt, id=id)
+        return Trajectory(x=x_arr, y=y_arr, dt=dt, traj_id=traj_id)
 
     def _export_trajectories(self, pix_per_m):
         t_list = []
@@ -724,7 +740,7 @@ class TrackingScenario():
             affine_params = np.array(self.camera_tracker.history)
             theta, tx, ty, _ = affine_params.T
             tx, ty = tx / pix_per_m, ty / pix_per_m
-            # invert axis
+            # Invert axis
             theta *= -1
             ty *= -1
             reference = theta, tx, ty
@@ -733,11 +749,11 @@ class TrackingScenario():
             t = self._tracker2trajectory(otrack, pix_per_m)
             if self.camera_tracker:
                 t = add_dynamic_reference(t, reference)
-            t_list.append(t)        
+            t_list.append(t)
         return t_list
 
     def track(self, video_path: str, start_in_frame: int = 0,
-              pix_per_m: int = 1):
+              end_in_frame: int = None, pix_per_m: int = 1):
         """
         Starts the tracking process.
 
@@ -746,21 +762,29 @@ class TrackingScenario():
         video_path : str
             Path of the video used to track the objects.
         start_in_frame : int, optional
-            Initial frame in which starts the processing. (Default is 0)
+            Initial frame in which starts the processing, by default 0.
+        end_in_frame : int, optional
+            Last frame being processed, if nothing is passed all frames
+            until the end of the video will be processed, by default
+            None.
         pix_per_m : int, optional
-            Pixel per meters. (Default is 1)
+            Pixel per meters, by default 1.
 
-            This value is used to readjuts the trajectories points to a real
-            scale.
+            This value is used to readjuts the trajectories points to a
+            real scale.
 
         Returns
         -------
         bool
             Whether or not the tracking process ended succefully.
         list
-            List of all the trajectories extracted in the tracking process.
+            List of all the trajectories extracted in the tracking
+            process.
         """
 
+        if end_in_frame is not None and end_in_frame > start_in_frame:
+            self.last_frame = int(end_in_frame)
+
         self._digest_video_path(video_path)
 
         end = False
